# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import json
from collections.abc import Iterator
from datetime import datetime
from pathlib import Path
from typing import Optional, Union

import click

try:
    import requests
except ImportError:  # pragma: no cov
    requests = None

from .__types import DictStr

cli_emoji: click.Command


# NOTE:
#   The Emoji dataset that provide by GitHub.
#   (refs: https://github.com/github/gemoji/blob/master/db/emoji.json)
GH_EMOJI_URL: str = (
    "https://raw.githubusercontent.com/github/gemoji/master/db/emoji.json"
)


def get_emojis() -> Iterator[DictStr]:
    """Get the iterator of the emoji data that already loading to assets path.
    This function use iterator for returning step because I do not want to keep
    all emoji data in memory.

    :rtype: Iterator[DictStr]
    """
    file = Path(__file__).parent / "assets/emoji.json"
    with file.open(encoding="utf-8") as f:
        yield from iter(json.load(f))


def demojize(
    msg: str,
    *,
    emojis: Optional[Union[Iterator[DictStr], list[DictStr]]] = None,
) -> str:
    """Replace a Unicode emoji to an emoji string in a message.

    :param msg: A message string that want to search emoji string.
    :param emojis: An iterator or list of mapping of emoji values.

    :rtype: str
    """
    emojis: Iterator[DictStr] = get_emojis() if emojis is None else emojis
    for emoji in emojis:
        if (e := emoji["emoji"]) in msg:
            msg = msg.replace(e, f':{emoji["alias"]}:')
    return msg


def emojize(
    msg: str,
    *,
    emojis: Optional[Union[Iterator[DictStr], list[DictStr]]] = None,
) -> str:
    """Replace an emoji string to a Unicode emoji in a message.

    :param msg: A message string that want to search Unicode emoji.
    :param emojis: An iterator or list of mapping of emoji values.

    :rtype: str
    """
    emojis: Iterator[DictStr] = get_emojis() if emojis is None else emojis
    for emoji in emojis:
        if (alias := f':{emoji["alias"]}:') in msg:
            msg = msg.replace(alias, emoji["emoji"])
    return msg


@click.group(name="emoji")
def cli_emoji():
    """The Emoji commands"""
    pass  # pragma: no cov


@cli_emoji.command()
@click.option("-b", "--backup", is_flag=True)
def fetch(backup: bool = False) -> None:
    """Refresh emoji metadata file on the assets folder, `./asserts`.

    :param backup: A backup flag for rename the previous file with backup suffix
        if this value set to True.
    :type backup: bool (=False)
    """
    if requests is None:  # pragma: no cov
        raise ImportError(
            "fetch command want the request package for getting the emoji "
            "metadata from the GitHub repository. Please install with: "
            "`pip install -U requests`"
        )

    file: Path = Path(__file__).parent / "assets/emoji.json"
    file.parent.mkdir(parents=True, exist_ok=True)

    if file.exists() and backup:
        file.rename(file.parent / f"emoji.bk{datetime.now():%Y%m%d%H%M%S}.json")

    with file.open(mode="w", encoding="utf-8") as f:
        json.dump(
            [
                {"emoji": data["emoji"], "alias": data["aliases"][0]}
                for data in requests.get(GH_EMOJI_URL).json()
            ],
            f,
            indent=2,
        )


if __name__ == "__main__":
    cli_emoji.main()



from datetime import datetime
from pathlib import Path
from unittest.mock import DEFAULT, patch

from click.testing import CliRunner

import clishelf.emoji as emoji


def side_effect_func(*args, **kwargs):
    if "emoji.py" in args[0]:
        _ = kwargs
        return Path(__file__)
    return DEFAULT


@patch("clishelf.emoji.Path", side_effect=side_effect_func)
@patch("clishelf.emoji.datetime")
@patch("clishelf.emoji.requests.get")
def test_fetch_emoji(mock_request, mock_now, mock_path):
    runner = CliRunner()

    mock_request.get.return_value.json.return_value = []
    mock_now.now.return_value = datetime(2024, 1, 1, 0, 0, 0)
    result = runner.invoke(emoji.fetch)
    assert mock_path.called
    assert 0 == result.exit_code

    test_file: Path = Path(__file__).parent / "assets/emoji.json"
    assert test_file.exists()

    result = runner.invoke(emoji.fetch, args="-b")
    assert result.exit_code == 0

    test_file_bk = Path(__file__).parent / "assets/emoji.bk20240101000000.json"
    assert test_file_bk.exists()

    test_file.unlink()
    test_file_bk.unlink()
    test_file_bk.parent.rmdir()



import datetime as dt
import subprocess
import sys
from unittest.mock import DEFAULT, patch

import clishelf.git as git
from clishelf.utils import Profile


def side_effect_func(*args, **kwargs):
    if any(["git", "rev-parse", "--abbrev-ref", "HEAD"] == a for a in args):
        _ = kwargs
        return "0.1.2".encode(encoding=sys.stdout.encoding)
    elif any(["git", "describe", "--tags", "--abbrev=0"] == a for a in args):
        _ = kwargs
        return "v0.0.1".encode(encoding=sys.stdout.encoding)
    else:
        return DEFAULT


def test_commit_message():
    msg = git.CommitMsg(content="test: test commit message", body="")
    assert ":test_tube: test: test commit message" == msg.content
    assert "Code Changes" == msg.mtype


def test_commit_log():
    commit_log = git.CommitLog(
        hash="",
        refs="",
        date=dt.datetime(2021, 1, 1),
        msg=git.CommitMsg(content="test: test commit message", body="|"),
        author=Profile(name="Demo Username", email="demo@mail.com"),
    )
    assert ":test_tube: test: test commit message" == commit_log.msg.content


@patch("clishelf.git.subprocess.check_output", side_effect=side_effect_func)
def test_get_latest_tag(mock):
    result = git.get_latest_tag()
    assert mock.called
    assert "v0.0.1" == result


@patch(
    "clishelf.git.subprocess.check_output",
    side_effect=subprocess.CalledProcessError(1, "git"),
)
def test_get_latest_tag_raise(mock):
    # Start Test after mock subprocess.
    result = git.get_latest_tag()
    assert mock.called
    assert "v0.0.0" == result



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import os
import re
import subprocess
import sys
from collections import defaultdict
from collections.abc import Iterator
from pathlib import Path
from typing import Any, NoReturn, Optional, TextIO, Union

import click

from .git import CommitLog
from .settings import BumpVerConf
from .utils import load_config

cli_vs: click.Command

GroupCommitLog = dict[str, list[CommitLog]]
TagGroupCommitLog = dict[str, GroupCommitLog]

UTF8: str = "utf-8"
HEAD: str = "HEAD"
CHANGELOG_HEADER: str = "# Changelogs"
CHANGELOG_LATEST: str = "## Latest Changes"
LINESEP: str = os.linesep


def map_group_commit_logs(
    *,
    all_tags: bool = False,
    is_dt: bool = False,
) -> TagGroupCommitLog:
    """Mapping Group to the getting commit logs function.

    :param all_tags: All tags flag.
    :param is_dt:

    :rtype: GroupCommitLog
    """
    from .git import get_commit_logs

    tag_group_logs: TagGroupCommitLog = defaultdict(lambda: defaultdict(list))
    for log in get_commit_logs(
        all_logs=all_tags,
        excluded=[
            r"pre-commit autoupdate",
            r"^Merge",
        ],
        is_dt=is_dt,
    ):
        tag_group_logs[log.refs][log.msg.mtype].append(log)

    rs: TagGroupCommitLog = {}
    for ref_tag in tag_group_logs:
        rs[ref_tag] = {
            k: sorted(v, key=lambda x: x.date, reverse=True)
            for k, v in tag_group_logs[ref_tag].items()
        }
    return rs


def get_changelog(
    file: Union[str, Path],
    *,
    tags: Optional[list[str]] = None,
    refresh: bool = False,
) -> Iterator[str]:
    """Return the list of content in the changelog file.

    :param file: A change log file path.
    :type file: str
    :param tags:
    :param refresh: A refresh flag

    :rtype: Iterator[str]
    """
    if refresh or not Path(file).exists():
        headers: list[str] = [CHANGELOG_HEADER, CHANGELOG_LATEST]
        if tags:
            headers.extend(f"## {t}" for t in tags)

        # NOTE: Add `""` to all spaces between rows.
        return iter(("|SEP||SEP|".join(headers)).split("|SEP|"))

    with Path(file).open(mode="r", encoding=UTF8) as f_changes:
        changes: list[str] = f_changes.read().splitlines()

    return iter(changes)


def write_group_log(
    writer: TextIO,
    group_logs: GroupCommitLog,
    tag_value: str,
) -> None:
    """Write a group log.

    :param writer:
    :param group_logs:
    :param tag_value:
    """
    from .git import get_commit_prefix_group

    linesep: str = LINESEP
    if not group_logs or any(
        cpt.name in group_logs for cpt in get_commit_prefix_group()
    ):
        linesep: str = f"{LINESEP}{LINESEP}"

    writer.write(f"## {tag_value}{linesep}")

    for group in (
        cpt for cpt in get_commit_prefix_group() if cpt.name in group_logs
    ):
        writer.write(f"### {group.emoji} {group.name}{LINESEP}{LINESEP}")
        for log in group_logs[group.name]:

            subject_fmt: str = (
                load_config()
                .get("version", {})
                .get("commit_subject_format", BumpVerConf.commit_subject_format)
            )
            subject: str = log.msg.subject.format(subject_fmt)

            msg_fmt: str = (
                load_config()
                .get("version", {})
                .get("commit_msg_format", BumpVerConf.commit_msg_format)
            )

            message: str = msg_fmt.format(
                subject=subject,
                datetime=log.date,
            )

            # NOTE: Write commit message to its refs.
            writer.write(f"{message}{LINESEP}")

        writer.write(LINESEP)


def create_changelog(
    file: Union[str, Path],
    *,
    all_tags: bool = False,
    refresh: bool = False,
    is_dt: bool = False,
) -> None:
    """Create Changelog file that generate from Git Log command.

    :param file:
    :param all_tags:
    :param refresh:
    :param is_dt:
    """
    group_logs: TagGroupCommitLog = map_group_commit_logs(
        all_tags=all_tags,
        is_dt=is_dt,
    )
    tags: list[str] = list(filter(lambda t: t != HEAD, group_logs.keys()))
    prev_change: Iterator[str] = get_changelog(file, tags=tags, refresh=refresh)

    with Path(file).open(mode="w", encoding=UTF8, newline="") as writer:
        skip_line: bool = False
        for line in prev_change:
            if line.startswith(CHANGELOG_LATEST):

                # NOTE: Start write group log for the HEAD refs.
                write_group_log(
                    writer,
                    group_logs.get(HEAD, {}),
                    tag_value="Latest Changes",
                )
                skip_line: bool = True

            elif m := re.match(rf"^##\s({BumpVerConf.get_regex(is_dt)})", line):
                get_tag: str = m.group(1)
                if get_tag in tags:  # pragma: no cov
                    write_group_log(
                        writer,
                        group_logs[get_tag],
                        tag_value=get_tag,
                    )
                    skip_line = True
                else:
                    skip_line = False
            elif line.startswith("## "):
                skip_line = False

            if not skip_line:
                writer.write(line + LINESEP)


def write_bump_file(
    param: dict[str, Any],
    *,
    version: int = 1,
    is_dt: bool = False,
) -> None:
    """Writing the ``.bump2version.cfg`` config file at current path.

    :param param:
    :param version:
    :param is_dt:
    """
    files: list[str] = load_config().get("version", {}).get("files", [])
    with Path(".bumpversion.cfg").open(mode="w", encoding=UTF8) as f:
        f.write(
            BumpVerConf.get_version(
                version,
                params=param,
                is_dt=is_dt,
            )
        )

        f.write("\n")
        for file in files:
            f.write(f"\n[bumpversion:file:{file}]\n")


def bump2version(
    action: str,
    file: str,
    changelog_file: str,
    changelog_ignore: bool = False,
    dry_run: bool = False,
    version: int = 1,
    *,
    is_dt: bool = False,
) -> None:  # pragma: no cov
    """Bump version processes that include:

        - write the bump2version file config on the current path
        - write changelog file if not ignore flag
        - commit the above steps to git
        - running bump2version cli for bump the version with the config
        - reset the latest commit
        - remove the bump2version file config
        - commit all change to the latest commit the running from bump2version

    :param action:
    :param file:
    :param changelog_file:
    :param changelog_ignore:
    :param dry_run:
    :param version:
    :param is_dt: A datetime mode flag
    """
    # Start writing ``.bump2version.cfg`` file on current path.
    click.echo("Start write '.bump2version.cfg' config file ...")
    write_bump_file(
        param={
            "changelog": changelog_file,
            "version": current_version(file, is_dt=is_dt),
            "action": action,
            "file": file,
        },
        version=version,
        is_dt=is_dt,
    )

    if not changelog_ignore:
        create_changelog(file=changelog_file)

    # COMMIT: commit add config and edit changelog file.
    subprocess.run(["git", "add", "-A"])
    subprocess.run(
        [
            "git",
            "commit",
            "-m",
            "build: add bump2version config file",
            "--no-verify",
        ],
        stdout=subprocess.DEVNULL,
    )
    click.echo("Running the `bump2version` cli with that config file ...")
    subprocess.run(
        [
            "bump2version",
            action,
            "--commit-args=--no-verify",
        ]
        + (["--list", "--dry-run"] if dry_run else [])
    )

    subprocess.run(
        [
            "git",
            "reset",
            "--soft",
            "HEAD~1",
        ],
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
    )

    # Remove ``.bump2version.cfg`` file.
    click.echo("Unlink '.bump2version.cfg' config file ...")
    Path(".bumpversion.cfg").unlink(missing_ok=False)

    with Path(".git/COMMIT_EDITMSG").open(encoding=UTF8) as f_msg:
        raw_msg = f_msg.read().splitlines()

    subprocess.run(["git", "add", "-A"], stderr=subprocess.DEVNULL)
    subprocess.run(
        [
            "git",
            "commit",
            "--amend",
            "-m",
            raw_msg[0],
            "--no-verify",
        ],
        stderr=subprocess.DEVNULL,
    )


def current_version(file: str, *, is_dt: bool = False) -> str:
    """Return the current version.

    :param file:
    :param is_dt:

    :rtype: str
    """
    with Path(file).open(encoding=UTF8) as f:
        if is_dt and (search_dt := re.search(BumpVerConf.regex_dt, f.read())):
            return search_dt[0]
        elif search := re.search(BumpVerConf.regex, f.read()):
            return search[0]
    raise NotImplementedError(f"{file} does not implement version value.")


@click.group(name="vs")
def cli_vs():
    """The Versioning commands."""
    pass  # pragma: no cov


@cli_vs.command()
@click.option("-f", "--file", type=click.Path(exists=True))
@click.option("-n", "--new", is_flag=True)
def changelog(
    file: Optional[str],
    new: bool = False,
) -> NoReturn:  # pragma: no cov
    """Make a changelog file that generate form previous commits."""
    if not file:
        file: str = (
            load_config().get("version", {}).get("changelog", None)
            or "CHANGELOG.md"
        )
    if new:
        create_changelog(file, all_tags=True, refresh=new)
        sys.exit(0)
    create_changelog(file, refresh=new)
    sys.exit(0)


@cli_vs.command()
@click.option(
    "-f",
    "--file",
    type=click.Path(exists=True),
    help="The contain version file that able to search with regex.",
)
def current(file: str) -> NoReturn:  # pragma: no cov
    """Return Current Version that read from ``__about__`` by default."""
    if not file:
        file: str = load_config().get("version", {}).get("version", None) or (
            f"./{load_config().get('project', {}).get('name', 'unknown')}"
            f"/__about__.py"
        )
    click.echo(current_version(file))
    sys.exit(0)


@cli_vs.command()
@click.argument(
    "action",
    type=click.STRING,
    required=1,
)
@click.option(
    "-f",
    "--file",
    type=click.Path(exists=True),
    help="A about file path that want to write new version.",
)
@click.option(
    "-c",
    "--changelog-file",
    type=click.Path(exists=True),
    help="A changelog file path that want to write new version.",
)
@click.option(
    "-m",
    "--mode",
    type=click.STRING,
    help="A bump version mode that should be normal or datetime.",
)
@click.option(
    "-v",
    "--version",
    type=click.INT,
    default=1,
    help="A version of bump2version config, it default by 1.",
)
@click.option(
    "--ignore-changelog",
    is_flag=True,
    help="If True, it will skip writing changelog step before bump version.",
)
@click.option(
    "--dry-run",
    is_flag=True,
    help="If True, it will pass --dry-run option to bump2version",
)
def bump(
    action: str,
    file: Optional[str] = None,
    changelog_file: Optional[str] = None,
    mode: Optional[str] = None,
    version: int = 1,
    ignore_changelog: bool = False,
    dry_run: bool = False,
) -> NoReturn:  # pragma: no cov
    """Bump package version with a next tag value with an input action.

    ACTION is the part of action that should be `major`, `minor`, or `patch`.

    \f
    :param action: A action path for bump the next version.
    :type action: str
    :param file:
    :type file: Optional[str] (=None)
    :param changelog_file: Optional[str] (=None)
    :param mode: Optional[str] (=None)
    :param version: int
    :param ignore_changelog: Ignore the changelog file if set be True.
    :type ignore_changelog: boolean
    :param dry_run: Dry run the bumpversion command if set be True.
    :type dry_run: boolean
    """
    vs_conf: dict[str, Any] = load_config().get("version", {})
    if not file:
        file: str = vs_conf.get("version", None) or (
            f"./{load_config().get('project', {}).get('name', 'unknown')}"
            f"/__about__.py"
        )
    if not changelog_file:
        changelog_file: str = vs_conf.get("changelog", None) or "CHANGELOG.md"
    if not mode:
        mode: str = vs_conf.get("mode", "normal")

    assert mode in (
        "datetime",
        "normal",
    ), "`mode` should be normal or datetime only"

    if mode == "normal":
        click.echo("Be noted that:")
        click.echo(
            "  * `major`:  means breaking changes and removed deprecations"
        )
        click.echo("  * `minor`:  new features, sometimes automatic migrations")
        click.echo("  * `patch`:  bug fixes")

    # NOTE: Start bumping version.
    bump2version(
        action,
        file,
        changelog_file,
        ignore_changelog,
        dry_run,
        version,
        is_dt=(mode == "datetime"),
    )
    sys.exit(0)


if __name__ == "__main__":
    cli_vs.main()



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import os
import re
import subprocess
import sys
from collections.abc import Iterator
from dataclasses import InitVar, dataclass, field
from datetime import date, datetime
from functools import lru_cache
from pathlib import Path
from typing import NoReturn, Optional

import click

from .__types import TupleStr
from .emoji import demojize, get_emojis
from .settings import GitConf
from .utils import (
    Level,
    Profile,
    load_config,
    make_color,
    prepare_str,
)

cli_git: click.Command

GIT_LOG_FORMAT: str = "%h|%D|%cI|%cn|%ce%n%s%n%b%-C()%n(END)"
DEFAULT_TAG: str = "v0.0.0"
ALL_CHAR: str = r"[\u0000-\uFFFF]"


def get_git_local_conf(key: str) -> Optional[str]:
    """Get Git config on the local scope with an input specific key.

    :param key: A key of config that want to get.
    :type key: str

    :rtype: Optional[str]
    """
    try:
        return (
            subprocess.check_output(["git", "config", "--local", key])
            .decode(sys.stdout.encoding)
            .strip()
        )
    except subprocess.CalledProcessError:  # pragma: no cover
        return None


def load_profile() -> Profile:
    """Load profile data from pyproject file or getting from git local config
    and return this values to Profile dataclass.

    :rtype: Profile
    """
    from .utils import load_pyproject

    _authors: dict[str, str] = (
        load_pyproject().get("project", {}).get("authors", {})
    )
    return Profile(
        name=_authors.get("name", get_git_local_conf("user.name")),
        email=_authors.get("email", get_git_local_conf("user.email")),
    )


@dataclass(frozen=True)
class CommitPrefix:
    """Commit prefix dataclass object that extract from a commit message."""

    name: str
    group: str
    emoji: str

    def __hash__(self) -> int:
        return hash(self.__str__())

    def __str__(self) -> str:
        return self.name


@dataclass(frozen=True)
class CommitPrefixGroup:
    """Commit prefix group dataclass object that was created from mapping of
    emoji group config.
    """

    name: str
    emoji: str

    def __hash__(self) -> int:
        return hash(self.__str__())

    def __str__(self) -> str:
        return self.name


def get_commit_prefix() -> Iterator[CommitPrefix]:
    """Return tuple of CommitPrefix that setting on the clishelf configuration.

    :rtype: Iterator[CommitPrefix]
    """
    # NOTE: Load config data that want to override.
    conf: list[str] = load_config().get("git", {}).get("commit_prefix", [])
    prefix_conf: TupleStr = tuple(c[0] for c in conf)

    yield from (
        CommitPrefix(name=n, group=g, emoji=e)
        for n, g, e in (
            *[p for p in GitConf.commit_prefix if p[0] not in prefix_conf],
            *conf,
        )
    )


def get_commit_prefix_group() -> tuple[CommitPrefixGroup, ...]:
    """Return tuple of CommitPrefixGroup

    :rtype: tuple[CommitPrefixGroup, ...]
    """
    conf: list[str] = (
        load_config().get("git", {}).get("commit_prefix_group", [])
    )
    prefix_conf: TupleStr = tuple(_[0] for _ in conf)
    return tuple(
        CommitPrefixGroup(name=n, emoji=e)
        for n, e in (
            *conf,
            *[
                p
                for p in GitConf.commit_prefix_group
                if p[0] not in prefix_conf
            ],
        )
    )


@lru_cache
def get_git_emojis() -> list[dict[str, str]]:
    """Return the list of mapping of Git emoji values.

    :rtype: list[dict[str, str]]
    """
    cm_prefix: TupleStr = tuple(p.emoji.strip(":") for p in get_commit_prefix())
    return [emojis for emojis in get_emojis() if emojis["alias"] in cm_prefix]


@dataclass
class CommitSub:
    """Commit Subject dataclass."""

    emoji: str
    prefix: str
    subject: str

    @classmethod
    def default(cls) -> CommitSub:
        return CommitSub(
            emoji=GitConf.commit_prefix_emoji_default,
            prefix="refactored",
            subject="default commit subject",
        )

    def format(self, fmt: str) -> str:
        return fmt.format(
            emoji=self.emoji,
            prefix=self.prefix,
            subject=self.subject,
        )


def extract_subject(content: str) -> CommitSub:
    """Extract string subject that receive from content string.

    :param content: A string content that want to extract.
    :type content: str

    :rtype: CommitSub
    """
    content: str = prepare_str(content)
    git_config: dict[str, bool] = load_config().get("git", {})

    if git_config.get("commit_prefix_pre_demojize", True):
        content: str = demojize(content)

    # NOTE: Fix case for the merge message that auto gen from git.
    if content.startswith("Merge branch "):
        return CommitSub(
            emoji=":fast_forward:",
            prefix="merge",
            subject=content.replace("Merge branch ", "branch "),
        )

    if rs := re.search(
        rf"^(?P<emoji>:\w+:)\s(?P<prefix>\w+):\s?(?P<subject>{ALL_CHAR}+)$",
        content,
    ):
        return CommitSub(**rs.groupdict())

    elif rs := re.search(
        rf"^(?P<emoji>:\w+:)\s(?P<subject>{ALL_CHAR}+)$",
        content,
    ):
        return CommitSub(prefix="refactored", **rs.groupdict())

    elif not (
        rs := re.search(
            rf"^(?P<prefix>\w+):\s?(?P<subject>{ALL_CHAR}+)$",
            content,
        )
    ):
        if not git_config.get("commit_prefix_force_fix", False):
            raise ValueError(
                f"The commit message does not support, {content!r}."
            )
        return CommitSub(
            emoji=":construction:",
            prefix="refactored",
            subject=content,
        )

    rs_dict: dict[str, str] = rs.groupdict()

    prefix: str = prepare_str(rs_dict["prefix"])
    content: str = prepare_str(rs_dict["subject"])

    emoji: Optional[str] = None
    for cp in get_commit_prefix():
        if prefix == cp.name:
            emoji: str = cp.emoji
            break

    if emoji is None and not git_config.get("commit_prefix_force_fix", False):
        raise ValueError(
            f"The prefix of this commit message does not support, "
            f"{prefix!r}."
        )

    return CommitSub(
        emoji=emoji or ":construction:",
        prefix=prefix,
        subject=content,
    )


@dataclass
class CommitMsg:
    """Commit Message dataclass that prepare un-emoji-prefix in that message."""

    content: InitVar[str]
    mtype: InitVar[Optional[str]] = field(default=None)
    body: str = field(default=None)  # NOTE: Mark new-line with ``|``
    subject: CommitSub = field(default_factory=CommitSub.default)

    def __str__(self) -> str:
        return f"{self.mtype}: {self.content}"

    def __post_init__(self, content: str, mtype: Optional[str] = None) -> None:
        """Post initialize dunder method on this dataclass for preparing
        the content and mtype fields.

        :param content:
        :param mtype:
        """
        self.subject: CommitSub = extract_subject(
            demojize(content, emojis=get_git_emojis())
        )
        self.content: str = self.__prepare_content__(self.subject)
        if mtype is None:
            self.mtype: str = self.__prepare_mtype(self.subject.prefix)
        else:
            self.mtype: str = mtype

    @staticmethod
    def __prepare_mtype(prefix: str) -> str:
        """Return a message type that getting from the regex.

        :rtype: str
        """
        for cp in get_commit_prefix():
            if prefix == cp.name:
                return cp.group

        return GitConf.commit_prefix_group_default

    @property
    def mtype_icon(self) -> str:
        return next(
            (
                cpt.emoji
                for cpt in get_commit_prefix_group()
                if cpt.name == self.mtype
            ),
            GitConf.commit_prefix_group_emoji_default,
        )  # pragma: no cover

    @staticmethod
    def __prepare_content__(subject: CommitSub) -> str:
        """Prepare string content that receive on post initialize step.

        :param subject: A CommitSub dataclass object.
        :type subject: CommitSub

        :rtype: str
        :return: A prepared string content that has an emoji prefix.
        """
        fmt: str = (
            load_config()
            .get("git", {})
            .get("commit_msg_format", GitConf.commit_msg_format)
        )
        return subject.format(fmt)


@dataclass(frozen=True)
class CommitLog:
    """Commit Log dataclass that use to keep commit log data from `git log` cli."""

    hash: str
    refs: str
    date: date
    msg: CommitMsg
    author: Profile

    def __str__(self) -> str:
        return "|".join(
            (
                self.hash,
                self.date.strftime("%Y-%m-%d"),
                self.msg.content,
                self.author.name,
                self.author.email,
                self.refs,
            )
        )


def _validate_commit_msg_warning(lines: list[str]) -> list[str]:
    """Validate Commit message that should to fixed, but it does not impact to
    target repository.

    :param lines: A list of line from commit message.
    :type lines: list[str]

    :rtype: list[str]
    :return: A list of warning message.
    """
    subject: str = lines[0]
    rs: list[str] = []

    # RULE 02: Limit the subject line to 50 characters
    if len(subject) <= 20 or len(subject) > 50:
        rs.append(
            "There should be between 21 and 50 characters in the commit title."
        )
    if len(lines) <= 2:
        rs.append("There should at least 3 lines in your commit message.")

    # RULE 01: Separate subject from body with a blank line
    if lines[1].strip() != "":
        rs.append(
            "There should be an empty line between the commit title and body."
        )

    if not lines[0].strip().endswith("."):
        lines[0] = f"{lines[0].strip()}."
        rs.append("There should not has dot in the end of commit message.")
    return rs


def validate_commit_msg(lines: list[str]) -> tuple[list[str], Level]:
    """Validate Commit message

    :param lines: A list of line from commit message.
    :type lines: List[str]

    :rtype: Tuple[List[str], Level]
    :return: A pair of warning messages and its logging level.
    """
    if not lines:
        return (
            ["Please supply commit message without start with ``#``."],
            Level.ERROR,
        )

    rs: list[str] = _validate_commit_msg_warning(lines)
    for line, msg in enumerate(lines[1:], start=2):
        # RULE 06: Wrap the body at 72 characters
        if len(msg) > 72:
            rs.append(
                f"The commit body should wrap at 72 characters at line: {line}."
            )

    if not rs:
        return (
            ["The commit message has the required pattern."],
            Level.OK,
        )
    return rs, Level.WARNING


def get_latest_tag(default: bool = True) -> Optional[str]:
    """Return the latest tag if it exists, otherwise it will v0.0.0 tag.

    :param default: A default flag that use DEFAULT_TAG value instead if it
        raises any error from subprocess process.

    :rtype: Optional[str]
    """
    try:
        return (
            subprocess.check_output(
                ["git", "describe", "--tags", "--abbrev=0"],
                stderr=subprocess.DEVNULL,
            )
            .decode(sys.stdout.encoding)
            .strip()
        )
    except subprocess.CalledProcessError:
        return DEFAULT_TAG if default else None


def gen_commit_logs(tag2head: str) -> Iterator[list[str]]:  # pragma: no cov
    """Prepare contents logs to List of commit log.

    :param tag2head: A length of log format string that want to get log from git
        log cli.

    :rtype: Iterator[list[str]]
    """
    prepare: list[str] = []
    for line in (
        subprocess.check_output(
            [
                "git",
                "log",
                tag2head,
                f"--pretty=format:{GIT_LOG_FORMAT}",
                "--date=iso8601-strict",
            ]
        )
        .decode(sys.stdout.encoding)
        .strip()
        .splitlines()
    ):
        # NOTE: Release log data if it found end line marking.
        if line == "(END)":
            yield prepare
            prepare = []
            continue

        prepare.append(line)


def get_commit_logs(
    tag: Optional[str] = None,
    *,
    all_logs: bool = False,
    excluded: Optional[list[str]] = None,
    is_dt: bool = False,
) -> Iterator[CommitLog]:  # pragma: no cov
    """Return a list of message that getting from commit log command.

    :param tag: A tag name that want to filter the commit log get to the HEAD.
    :type tag: Optional[str] (=None)
    :param all_logs: A flag that make this function get all logs.
    :type all_logs: bool (=False)
    :param excluded: A list of excluded regular expression string.
    :type excluded: Optional[list[str]] (=None)
    :param is_dt: A datetime mode flag.
    :type is_dt: bool(=False)

    :rtype: Iterator[CommitLog]
    """
    from .settings import BumpVerConf

    # NOTE: Prepare tag to head value for getting Git logs.
    if tag:
        tag2head: str = f"{tag}..HEAD"
    elif all_logs or not (tag := get_latest_tag(default=False)):
        tag2head = "HEAD"
    else:
        tag2head = f"{tag}..HEAD"

    refs: str = "HEAD"
    for logs in gen_commit_logs(tag2head):

        if any(
            re.search(s, logs[1]) is not None for s in (excluded or [r"^Merge"])
        ):
            continue

        header: list[str] = logs[0].split("|")
        if ref_tag := [
            ref.strip() for ref in header[1].strip().split(",") if "tag:" in ref
        ]:
            if search := re.search(
                rf"tag:\sv?(?P<version>{BumpVerConf.get_regex(is_dt)})",
                ref_tag[0],
            ):
                refs = search.groupdict()["version"]

        yield CommitLog(
            hash=header[0],
            refs=refs,
            date=datetime.fromisoformat(header[2]),
            msg=CommitMsg(
                content=logs[1],
                body="|".join(logs[2:]),
            ),
            author=Profile(
                name=header[3],
                email=header[4],
            ),
        )


def merge2latest_commit(no_verify: bool = False) -> None:  # pragma: no cov
    """Merge all stage changes to the previous commit with the same commit
    message.
    """
    subprocess.run(
        ["git", "commit", "--amend", "--no-edit", "-a"]
        + (["--no-verify"] if no_verify else [])
    )


def get_latest_commit(
    file: Optional[str] = None,
    edit: bool = False,
    output_file: bool = False,
) -> list[str]:  # pragma: no cov
    """Return a list of line that created on commit message file.

    :param file:
    :param edit:
    :param output_file:

    :rtype: list[str]
    """
    if file:
        with Path(file).open(encoding="utf-8") as f_msg:
            raw_msg = f_msg.read().splitlines()
    else:
        raw_msg = (
            subprocess.check_output(
                ["git", "log", "HEAD^..HEAD", "--pretty=format:%B"]
            )
            .decode(sys.stdout.encoding)
            .strip()
            .splitlines()
        )
    lines: list[str] = [
        msg for msg in raw_msg if not msg.strip().startswith("#")
    ]
    if lines[-1] != "":
        lines += [""]  # Add end-of-file line

    rss, level = validate_commit_msg(lines)
    for rs in rss:
        click.echo(make_color(rs, level))
    if level not in (Level.OK, Level.WARNING):
        sys.exit(1)

    if edit:
        lines[0] = CommitMsg(content=lines[0]).content

    if file and output_file:
        with Path(file).open(mode="w", encoding="utf-8", newline="") as f_msg:
            f_msg.write(f"{os.linesep}".join(lines))
    return lines


@click.group(name="git")
def cli_git():
    """The Extended Git commands"""
    pass  # pragma: no cov


@cli_git.command()
@click.argument(
    "file",
    type=click.STRING,
    default=".git/COMMIT_EDITMSG",
)
@click.option("-e", "--edit", is_flag=True)
@click.option("-o", "--output-file", is_flag=True)
@click.option("-p", "--prepare", is_flag=True)
def cm(
    file: Optional[str],
    edit: bool,
    output_file: bool,
    prepare: bool,
) -> None:  # pragma: no cov
    """Prepare and show the latest commit message with the commit message
    general rules.

    \f
    :param file:
    :param edit:
    :param output_file:
    :param prepare:
    """
    if not prepare:
        click.echo(
            make_color(
                "\n".join(get_latest_commit(file, edit, output_file)),
                level=Level.OK,
            ),
        )
    else:
        edit: bool = True
        _cm_msg: str = "\n".join(get_latest_commit(file, edit, output_file))
        subprocess.run(
            [
                "git",
                "commit",
                "--amend",
                "-a",
                "--no-verify",
                "-m",
                _cm_msg,
            ],
            stdout=subprocess.DEVNULL,
        )
        click.echo(make_color(_cm_msg, level=Level.OK))
    sys.exit(0)


@cli_git.command()
@click.option("-g", "--group", is_flag=True)
def cm_msg(group: bool = False) -> None:  # pragma: no cov
    """Return list of commit prefixes"""
    if group:
        for cm_prefix_g in get_commit_prefix_group():
            click.echo(f"{cm_prefix_g.emoji} {cm_prefix_g.name}")
    else:
        for cm_prefix in get_commit_prefix():
            click.echo(
                f"{cm_prefix.emoji} {cm_prefix.name} -> {cm_prefix.group}"
            )
    sys.exit(0)


@cli_git.command()
@click.option("--verify", is_flag=True)
def cm_prev(verify: bool = False) -> None:  # pragma: no cov
    """Commit changes to the Previous Commit with same message.

    \f
    :param verify: A verify flag before commit.
    """
    merge2latest_commit(no_verify=(not verify))
    sys.exit(0)


@cli_git.command()
@click.option("-f", "--force", is_flag=True)
@click.option("-n", "--number", type=click.INT, default=1)
def cm_revert(force: bool, number: int = 1) -> None:  # pragma: no cov
    """Revert the latest Commit on the Local repository.

    \f
    :param force: A force flag that restore and clean all stage after reset.
    :param number: A number of commit that want to revert from the HEAD.
    """
    subprocess.run(["git", "reset", f"HEAD~{number}"])
    if force:
        subprocess.run(["git", "restore", "."])
        subprocess.run(["git", "clean", "-f"])
    sys.exit(0)


@cli_git.command()
@click.argument(
    "branch",
    type=click.STRING,
    default="main",
)
@click.option(
    "-t",
    "--theirs",
    is_flag=True,
    help="If True, it will use `their` strategy if it has conflict",
)
@click.option(
    "-o",
    "--ours",
    is_flag=True,
    help="If True, it will use `ours` strategy if it has conflict",
)
@click.option(
    "-s",
    "--squash",
    is_flag=True,
    help="If True, it will use `squash` merge option.",
)
def mg(
    branch: str,
    theirs: bool = False,
    ours: bool = False,
    squash: bool = False,
) -> None:  # pragma: no cov
    """Merge change from another branch with strategy, `theirs` or `ours`.

    BRANCH is a name of branch that you want to merge with current branch.

    \f
    :param branch: A name of branch that you want to merge with current branch.
    :param theirs: If True, it will use `their` strategy if it has conflict.
    :param ours: If True, it will use `ours` strategy if it has conflict.
    :param squash: If True, it will use `squash` merge option.
    """
    if theirs and ours:
        raise ValueError("The strategy flag should not True together.")
    elif ours:
        strategy = "ours"
    else:
        strategy = "theirs"
    subprocess.run(
        [
            "git",
            "merge",
            branch,
            "--strategy-option",
            strategy,
            *(["--squash"] if squash else []),
        ],
        stderr=subprocess.DEVNULL,
    )
    sys.exit(0)


@cli_git.command()
def bn_clear() -> NoReturn:  # pragma: no cov
    """Clear Local Branches that sync from the Remote repository."""
    subprocess.run(
        ["git", "checkout", "main"],
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
    )
    subprocess.run(
        # Or, use ``git remote prune origin``.
        ["git", "remote", "update", "origin", "--prune"],
        stdout=subprocess.DEVNULL,
    )
    branches = (
        subprocess.check_output(["git", "branch", "-vv"])
        .decode(sys.stdout.encoding)
        .strip()
        .splitlines()
    )
    for branch in branches:
        if ": gone]" in branch:
            subprocess.run(["git", "branch", "-D", branch.strip().split()[0]])
    subprocess.run(["git", "checkout", "-"])
    sys.exit(0)


@cli_git.command()
@click.option(
    "-p",
    "--push",
    is_flag=True,
    help="If True, it will auto push to remote",
)
def tg_bump(push: bool = False) -> None:  # pragma: no cov
    """Create Tag from current version after bumping

    \f
    :param push: A push flag that will push a tag to remote if it set to True.
    """
    latest_tag: str = get_latest_tag(default=False)
    subprocess.run(
        ["git", "tag", "-d", f"{latest_tag}"],
        stderr=subprocess.DEVNULL,
    )
    subprocess.run(
        ["git", "fetch", "--prune", "--prune-tags"],
        stdout=subprocess.DEVNULL,
    )
    subprocess.run(["git", "tag", f"{latest_tag}"])
    if push:
        subprocess.run(["git", "push", f"{latest_tag}", "--tags"])
    sys.exit(0)


@cli_git.command()
def tg_clear() -> None:  # pragma: no cov
    """Clear Local Tags that sync from the Remote repository."""
    subprocess.run(
        ["git", "fetch", "--prune", "--prune-tags"],
        stdout=subprocess.DEVNULL,
    )
    sys.exit(0)


if __name__ == "__main__":
    cli_git.main()
